# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

import numpy as np


class BaseState(ABC):
    def __init__(self, state_fp, read_only=False):
        """Initialize State instance.

        state_fp: str
            Path to state file.
        read_only: bool
            Whether to open file in read only mode.
        """
        self.state_fp = state_fp
        self.read_only = read_only
        self.restore(state_fp)

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def __str__(self):
        return str(self.to_dict())

    @abstractmethod
    def set_labels(self, y):
        """Add/set labels to state

        If the labels do not exist, add it to the state.

        Arguments
        ---------
        y: numpy.ndarray
            One dimensional integer numpy array with inclusion labels.
        """
        raise NotImplementedError

    @abstractmethod
    def set_final_labels(self, y):
        """Add/set final labels to state.

        If final_labels does not exist yet, add it.

        Arguments
        ---------
        y: numpy.ndarray
            One dimensional integer numpy array with final inclusion labels.
        """
        raise NotImplementedError

    @abstractmethod
    def _add_as_data(self, as_data, feature_matrix=None):
        """Add properties from as_data to the state.

        Arguments
        ---------
        as_data: ASReviewData
            Data file from which the review is run.
        feature_matrix: np.ndarray, sklearn.sparse.csr_matrix
            Feature matrix computed by the feature extraction model.
        """
        raise NotImplementedError

    @abstractmethod
    def get_feature_matrix(self, data_hash):
        """Get feature matrix out of the state.

        Arguments
        ---------
        data_hash: str
            Hash of as_data object from which the matrix is derived.

        Returns
        -------
        np.ndarray, sklearn.sparse.csr_matrix:
            Feature matrix as computed by the feature extraction model.
        """
        raise NotImplementedError

    @abstractmethod
    def get_current_queries(self):
        """Get the current queries made by the model.

        This is useful to get back exactly to the state it was in before
        shutting down a review.

        Returns
        -------
        dict:
            The last known queries according to the state file.
        """
        raise NotImplementedError

    @abstractmethod
    def set_current_queries(self, current_queries):
        """Set the current queries made by the model.

        Arguments
        ---------
        current_queries: dict
            The last known queries, with {query_idx: query_method}.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def settings(self):
        """Get settings from state
        """
        raise NotImplementedError

    @abstractmethod
    def add_classification(self, idx, labels, methods, query_i):
        """Add training indices and their labels.

        Arguments
        ---------
        indices: list, numpy.ndarray
            A list of indices used for training.
        labels: list
            A list of labels corresponding with the training indices.
        i: int
            The query number.
        """
        raise NotImplementedError

    @abstractmethod
    def add_proba(self, pool_idx, train_idx, proba, query_i):
        """Add inverse pool indices and their labels.

        Arguments
        ---------
        indices: list, numpy.ndarray
            A list of indices used for unlabeled pool.
        pred: numpy.ndarray
            Array of prediction probabilities for unlabeled pool.
        i: int
            The query number.
        """
        raise NotImplementedError

    def is_empty(self):
        """Check if state has no results.

        Returns
        -------
        bool
            True if empty.
        """
        return self.n_queries() == 0

    @abstractmethod
    def n_queries(self):
        """Number of queries saved in the state.

        Returns
        -------
        int
            Number of queries.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, variable, query_i=None, default=None, idx=None):
        """Get data from the state object.

        This is universal accessor method of the State classes. It can be used
        to get a variable from one specific query. In theory, it should get the
        whole data set if query_i=None, but this is not currently implemented
        in any of the States.

        Arguments
        ---------
        variable: str
            Name of the variable/data to get. Options are:
            label_idx, inclusions, label_methods, labels, final_labels, proba
            , train_idx, pool_idx.
        query_i: int
            Query number, should be between 0 and self.n_queries().
        idx: int, numpy.ndarray,list
            Indices to get in the returned array.
        """
        raise NotImplementedError

    @abstractmethod
    def delete_last_query(self):
        """Delete the last query from the state object."""
        raise NotImplementedError

    def startup_vals(self):
        """Get variables for reviewer to continue review.

        Returns
        -------
        numpy.ndarray:
            Current labels of dataset.
        numpy.ndarray:
            Current training indices.
        dict:
            Dictionary containing the sources of the labels.
        query_i:
            Currenty query number (starting from 0).
        """
        labels = self.get("labels")

        train_idx = []
        query_src = {}
        for query_i in range(self.n_queries()):
            try:
                label_idx = self.get("label_idx", query_i)
                labelled = self.get("inclusions", query_i)
                label_methods = self.get("label_methods", query_i)
            except (KeyError, IndexError):
                continue

            for i, meth in enumerate(label_methods):
                if meth not in query_src:
                    query_src[meth] = []
                query_src[meth].append(label_idx[i])
                labels[label_idx[i]] = labelled[i]
            train_idx.extend(label_idx)

        if query_i > 0:
            n_queries = self.n_queries()
            last_inclusions = None
            try:
                last_inclusions = self.get("inclusions", n_queries - 1)
            except KeyError:
                last_inclusions = []
            if last_inclusions is None:
                last_inclusions = []
            query_i_classified = len(last_inclusions)
        else:
            query_i_classified = 0

        train_idx = np.array(train_idx, dtype=np.int)
        startup_vals = {
            "labels": labels,
            "train_idx": np.unique(train_idx),
            "query_src": query_src,
            "query_i": query_i,
            "query_i_classified": query_i_classified,
        }
        return startup_vals

    def review_state(self):
        startup = self.startup_vals()
        return (startup["labals"], startup["train_idx"], startup["query_src"],
                startup["query_i"])

    @property
    def pred_proba(self):
        """Get last predicted probabilities."""
        for query_i in reversed(range(self.n_queries())):
            try:
                proba = self.get("proba", query_i=query_i)
                if proba is not None:
                    return proba
            except KeyError:
                pass
        return None

    @abstractmethod
    def initialize_structure(self):
        """Create empty internal structure for state"""
        raise NotImplementedError

    @abstractmethod
    def close(self):
        """Close the files opened by the state.

        Also sets the end time if not in read-only mode.
        """
        raise NotImplementedError

    @abstractmethod
    def save(self):
        """Save state to file.

        Arguments
        ---------
        fp: str
            The file path to export the results to.

        """
        raise NotImplementedError

    @abstractmethod
    def restore(self, fp):
        """Restore or create state from a state file.

        If the state file doesn't exist, creates and empty state that is ready
        for storage.

        Arguments
        ---------
        fp: str
            Path to file to restore/create.
        """
        raise NotImplementedError

    def to_dict(self):
        """Convert state to dictionary.

        Returns
        -------
        dict:
            Dictionary with all relevant variables.
        """
        state_dict = {}
        state_dict["settings"] = vars(self.settings)

        global_datasets = ["labels", "final_labels"]
        for dataset in global_datasets:
            try:
                state_dict[dataset] = self.get(dataset).tolist()
            except KeyError:
                pass

        query_datasets = [
            "label_methods", "label_idx", "inclusions", "proba", "pool_idx",
            "train_idx"
        ]
        state_dict["results"] = []
        for query_i in range(self.n_queries()):
            state_dict["results"].append({})
            for dataset in query_datasets:
                try:
                    state_dict["results"][query_i][dataset] = self.get(
                        dataset, query_i).tolist()
                except (KeyError, IndexError):
                    pass
        return state_dict



# Copyright 2019-2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
from io import BytesIO
from urllib.request import urlopen

from asreview.webapp.tests.utils import retrieve_project_url_github


# Retrieve urls to .asreview files exported from previous versions
project_urls = retrieve_project_url_github()


def test_retrieve_project_file():
    """Test retrieve project files from GitHub"""

    if not project_urls:
        assert False


@pytest.mark.parametrize("url", project_urls)
def test_project_file(tmp_path, client, url):
    """Test import and continue a project created in previous versions."""

    # change default folder for projects
    os.environ["ASREVIEW_PATH"] = str(tmp_path)

    # Test import uploaded project
    with urlopen(url) as project_file:
        response_import = client.post("/api/project/import_project", data={
            "file": (BytesIO(project_file.read()), "project.asreview")
        })
    json_data_import = response_import.get_json()
    assert response_import.status_code == 200

    project_path = f"/api/project/{json_data_import['id']}"

    # Test get projects
    response_projects = client.get("/api/projects")
    json_data_projects = response_projects.get_json()
    assert "result" in json_data_projects
    assert any(
        item["id"] == f"{json_data_import['id']}"
        for item in json_data_projects["result"]
    )

    # Test get info on the project
    response = client.get(f"{project_path}/info")
    json_data = response.get_json()
    assert json_data["id"] == json_data_import['id']

    # Test get progress info on the article
    response_progress = client.get(f"{project_path}/progress")
    json_data_progress = response_progress.get_json()
    assert isinstance(json_data_progress, dict)

    # Test get progress history on the article
    response_progress_history = client.get(f"{project_path}/progress_history")
    json_data_progress_history = response_progress_history.get_json()
    assert isinstance(json_data_progress_history, list)

    # Test get cumulative number of inclusions by ASReview/at random
    response_progress_efficiency = client.get(f"{project_path}/progress_efficiency")
    json_data_progress_efficiency = response_progress_efficiency.get_json()
    assert isinstance(json_data_progress_efficiency, list)

    # Test retrieve documents in order of review
    response_get_document = client.get(f"{project_path}/get_document")
    json_data_get_document = response_get_document.get_json()
    assert "result" in json_data_get_document
    assert isinstance(json_data_get_document, dict)

    # Test retrieve classification result
    response_classify_instance = client.post(f"{project_path}/record/<doc_id>", data={
        "doc_id": 99,
        "label": 1,
    })
    assert response_classify_instance.status_code == 200

    # Test update classification result
    response_update_classify = client.put(f"{project_path}/record/<doc_id>", data={
        "doc_id": 99,
        "label": 0,
    })
    assert response_update_classify.status_code == 200

    # Test retrieve review history
    response_prior = client.get(f"{project_path}/prior")
    json_data_prior = response_prior.get_json()
    assert "result" in json_data_prior
    assert isinstance(json_data_prior["result"], list)

    # Test export result
    response_export_result_excel = client.get(f"{project_path}/export?file_type=excel")
    response_export_result_csv = client.get(f"{project_path}/export?file_type=csv")
    response_export_result_tsv = client.get(f"{project_path}/export?file_type=tsv")
    assert response_export_result_excel.status_code == 200
    assert response_export_result_csv.status_code == 200
    assert response_export_result_tsv.status_code == 200

    # Test export project
    response_export_project = client.get(f"{project_path}/export_project")
    assert response_export_project.status_code == 200

    # Test finish project
    response_finish = client.get(f"{project_path}/finish")
    assert response_finish.status_code == 200

    # Test delete project
    response_delete = client.delete(f"{project_path}/delete")
    assert response_delete.status_code == 200



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from asreview.webapp.start_flask import create_app


@pytest.fixture
def app():

    app = create_app()

    return app


@pytest.fixture
def client(app):
    """A test client for the app."""
    return app.test_client()



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.datasets import BaseVersionedDataSet
from asreview.datasets import DatasetManager
from asreview.search import fuzzy_find
from asreview.utils import is_iterable
from asreview.webapp.utils.io import read_data


def search_data(project_id, q, n_max=100):
    """Get the title/authors/abstract for a paper."""

    # read the dataset
    as_data = read_data(project_id)

    # search for the keywords
    result_idx = fuzzy_find(
        as_data, q, max_return=n_max, exclude=[], by_index=True)

    # return full information on the records
    return as_data.record(result_idx, by_index=True)


def get_data_statistics(project_id):
    """Get the title/authors/abstract for a paper."""

    # read the dataset
    as_data = read_data(project_id)

    result = {
        "n_rows": as_data.df.shape[0],
        "n_cols": as_data.df.shape[1],
    }

    # return full information on the records
    return result


def get_dataset_metadata(exclude=None, include=None):

    manager = DatasetManager()
    groups = manager.groups.copy()

    if exclude is not None:

        # make iterable if not the case
        if not is_iterable(exclude):
            exclude = [exclude]

        # pop items
        for group_id in exclude:
            try:
                groups.remove(group_id)
            except ValueError:
                pass

    if include is not None:

        # make iterable if not the case
        if not is_iterable(include):
            include = [include]

        # pop items
        for group_id in groups.copy():
            if group_id not in include:
                groups.remove(group_id)

    # get datasets
    all_datasets = manager.list(
        group_name=groups,
        latest_only=False,
        raise_on_error=True
    )

    result_datasets = []
    for group_id, data_list in all_datasets.items():
        for dataset in data_list:
            if isinstance(dataset, BaseVersionedDataSet):
                cur_data = []
                for vdata in dataset.datasets:
                    vdata.dataset_id = f"{group_id}:{vdata.dataset_id}"
                    cur_data.append(vdata.to_dict())
                result_datasets.append(cur_data)
            else:
                dataset.dataset_id = f"{group_id}:{dataset.dataset_id}"
                result_datasets.append(dataset.to_dict())

    return result_datasets



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pandas as pd
import rispy
from rispy import TAG_KEY_MAPPING
from rispy.config import LIST_TYPE_TAGS

from asreview.config import COLUMN_DEFINITIONS
from asreview.io.utils import standardize_dataframe


RIS_KEY_LABEL_INCLUDED = "LI"


def _tag_key_mapping(reverse=False):
    # Add label_included into the specification and create reverse mapping.
    TAG_KEY_MAPPING[RIS_KEY_LABEL_INCLUDED] = "included"
    KEY_TAG_MAPPING = {TAG_KEY_MAPPING[key]: key for key in TAG_KEY_MAPPING}
    for label in COLUMN_DEFINITIONS["included"]:
        KEY_TAG_MAPPING[label] = "LI"
    if reverse:
        return KEY_TAG_MAPPING
    else:
        return TAG_KEY_MAPPING


def read_ris(fp):
    """RIS file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the RIS file.
    label: bool
        Check for label. If None, this is automatic.

    Returns
    -------
    pandas.DataFrame:
        Dataframe with entries.

    """

    encodings = ['ISO-8859-1', 'utf-8', 'utf-8-sig']
    entries = None
    for encoding in encodings:
        try:
            with open(fp, 'r', encoding=encoding) as bibliography_file:
                mapping = _tag_key_mapping(reverse=False)
                entries = list(rispy.load(bibliography_file, mapping=mapping))
                break
        except UnicodeDecodeError:
            pass
        except IOError as e:
            logging.warning(e)

    if entries is None:
        raise ValueError("Cannot find proper encoding for data file.")

    df = pd.DataFrame(entries)

    def converter(x):
        try:
            return ", ".join(x)
        except TypeError:
            return ""

    for tag in LIST_TYPE_TAGS:
        key = TAG_KEY_MAPPING[tag]
        if key in df:
            df[key] = df[key].apply(converter)
    return standardize_dataframe(df)


def write_ris(df, ris_fp):
    """Write dataframe to RIS file.

    Arguments
    ---------
    df: pandas.Dataframe
        Dataframe to export.
    ris_fp: str
        RIS file to export to.
    """
    column_names = list(df)
    column_key = []
    for col in column_names:
        try:
            rev_mapping = _tag_key_mapping(reverse=True)
            column_key.append(rev_mapping[col])
        except KeyError:
            column_key.append('UK')
            logging.info(f"Cannot find column {col} in specification.")

    n_row = df.shape[0]

    # According to RIS specifications, a record should begin with TY.
    # Thus, the column id is inserted before all the other.
    col_order = []
    for i, key in enumerate(column_key):
        if key == 'TY':
            col_order.insert(0, i)
        else:
            col_order.append(i)

    with open(ris_fp, "w") as fp:
        for i_row in range(n_row):
            for i_col in col_order:
                value = df.iloc[i_row, i_col]
                if isinstance(value, list):
                    for val in value:
                        fp.write(f"{column_key[i_col]}  - {val}\n")
                else:
                    fp.write(f"{column_key[i_col]}  - {value}\n")
            fp.write("ER  - \n\n")



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from asreview.io.utils import standardize_dataframe


def read_csv(data_fp):
    """CVS file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the CSV file.

    Returns
    -------
    list:
        List with entries.

    """

    for encoding in ["utf-8", "ISO-8859-1"]:
        try:
            df = pd.read_csv(
                data_fp,
                sep=None,
                encoding=encoding,
                engine='python'
            )
            return standardize_dataframe(df)
        except UnicodeDecodeError:
            # if unicode error, go to next encoding
            continue

    raise UnicodeDecodeError("The encoding of the file is not supported.")



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from asreview.dataset import ASReviewData
from asreview.webapp.utils.paths import get_project_path


def check_dataset(fp):

    # try:
    #     ASReviewData.from_file(fp)
    # except Exception as err:
    #     raise Exception("Incorrect file format")
    pass


def is_project(project_id):

    if get_project_path(project_id).exists():
        return True

    return False



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sqlite3
from time import sleep


def _log_msg(msg, project_id=None):
    """Return the log message with project_id."""
    if project_id is not None:
        return f"Project {project_id} - {msg}"

    return msg


def get_db(db_file):
    db = sqlite3.connect(str(db_file), detect_types=sqlite3.PARSE_DECLTYPES)
    db.row_factory = sqlite3.Row
    return db


def release_all_locks(db_file):
    db = get_db(db_file)
    db.execute('DELETE FROM locks;')
    db.close()


class SQLiteLock():
    def __init__(self,
                 db_file,
                 lock_name="global",
                 blocking=False,
                 timeout=30,
                 polling_rate=0.4,
                 project_id=None):
        self.db_file = db_file
        self.lock_name = lock_name
        self.lock_acquired = False
        self.timeout = timeout
        self.polling_rate = polling_rate
        self.project_id = project_id

        # acquire
        self.acquire(
            blocking=blocking, timeout=timeout, polling_rate=polling_rate)

    def acquire(self, blocking=False, timeout=30, polling_rate=0.4):
        if self.lock_acquired:
            return

        if not os.path.isfile(self.db_file):
            self.init_db()

        cur_timeout = 0
        while True and not self.lock_acquired:
            db = get_db(self.db_file)
            try:
                db.isolation_level = 'EXCLUSIVE'
                db.execute('BEGIN EXCLUSIVE')
                lock_entry = db.execute('SELECT * FROM locks WHERE name = ?',
                                        (self.lock_name, )).fetchone()
                if lock_entry is None:
                    db.execute('INSERT INTO locks (name) VALUES (?)',
                               (self.lock_name, ))
                    self.lock_acquired = True
                    logging.debug(
                        _log_msg(f"Acquired lock {self.lock_name}",
                                 self.project_id))
                db.commit()
            except sqlite3.OperationalError as e:
                logging.error(
                    _log_msg(f"Encountering operational error {e}",
                             self.project_id))
            db.close()
            if self.lock_acquired or not blocking:
                break
            cur_timeout += polling_rate
            sleep(polling_rate)

    def init_db(self):
        db = get_db(self.db_file)
        db.executescript('DROP TABLE IF EXISTS locks; '
                         'CREATE TABLE locks (name TEXT NOT NULL);')
        db.close()

    def locked(self):
        return self.lock_acquired

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.release()

    def release(self):
        if not self.locked():
            return
        while True:
            db = get_db(self.db_file)
            try:
                db.execute('DELETE FROM locks WHERE name = ?',
                           (self.lock_name, ))
                db.commit()
                db.close()
                break
            except sqlite3.OperationalError:
                pass
            db.close()
            sleep(0.4)
        logging.debug(
            _log_msg(f"Released lock {self.lock_name}", self.project_id))
        self.lock_acquired = False



# Copyright 2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from asreview.data.base import ASReviewData
from asreview.datasets import DataSetNotFoundError
from asreview.datasets import DatasetManager
from asreview.utils import is_url


def load_data(name, *args, **kwargs):
    """Load data from file, URL, or plugin.

    Parameters
    ----------
    name: str, pathlib.Path
        File path, URL, or alias of extension dataset.

    Returns
    -------
    asreview.ASReviewData:
        Inititalized ASReview data object.
    """

    # check is file or URL
    if Path(name).exists() or is_url(name):
        return ASReviewData.from_file(name, *args, **kwargs)

    # check if dataset is plugin dataset\
    try:
        dataset_path = DatasetManager().find(name).get()
        return ASReviewData.from_file(dataset_path, *args, **kwargs)
    except DataSetNotFoundError:
        pass

    # Could not find dataset, return None.
    raise FileNotFoundError(
        f"File, URL, or dataset does not exist: '{name}'")



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.io.utils import standardize_dataframe

import pandas as pd
from asreview.config import COLUMN_DEFINITIONS


def read_excel(fp):
    """Excel file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the Excel file (.xlsx).

    Returns
    -------
    list:
        List with entries.

    """
    try:
        dfs = pd.read_excel(fp, sheet_name=None)
    except UnicodeDecodeError:
        dfs = pd.read_excel(fp, sheet_name=None, encoding="ISO-8859-1")

    best_sheet = None
    sheet_obj_val = -1
    wanted_columns = []
    for type_name, type_list in COLUMN_DEFINITIONS.items():
        wanted_columns.extend(type_list)

    for sheet_name in dfs:
        col_names = set([col.lower() for col in list(dfs[sheet_name])])
        obj_val = len(col_names & set(wanted_columns))
        if obj_val > sheet_obj_val:
            sheet_obj_val = obj_val
            best_sheet = sheet_name

    return standardize_dataframe(dfs[best_sheet])



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import pytest


def test_get_projects(client):
    """Test get projects."""

    response = client.get("/api/projects")
    json_data = response.get_json()

    assert "result" in json_data
    assert isinstance(json_data["result"], list)


def test_init_project(tmp_path, client):
    """Test create project."""

    # change default folder for projects
    os.environ["ASREVIEW_PATH"] = str(tmp_path)

    response = client.post("/api/project/info", data={
        "name": "project id",
        "authors": "name",
        "description": "hello world"
    })
    json_data = response.get_json()

    assert response.status_code == 201
    assert "name" in json_data
    assert isinstance(json_data, dict)


def test_demo_data_project(client):
    """Test retrieve plugin and benchmark datasets"""

    response_plugin = client.get("/api/datasets?subset=plugin")
    response_benchmark = client.get("/api/datasets?subset=benchmark")
    json_plugin_data = response_plugin.get_json()
    json_benchmark_data = response_benchmark.get_json()

    assert "result" in json_plugin_data
    assert "result" in json_benchmark_data
    assert isinstance(json_plugin_data["result"], list)
    assert isinstance(json_benchmark_data["result"], list)


def test_upload_data_to_project(client):
    """Test add data to project."""

    response = client.post("/api/project/project-id/data", data={
        "benchmark": "benchmark:Hall_2012"
    })
    assert response.status_code == 200


def test_get_project_data(client):
    """Test get info on the data"""

    response = client.get("/api/project/project-id/data")
    json_data = response.get_json()
    assert json_data["filename"] == "Hall_2012"


def test_update_project_info(client):
    """Test update project info"""

    response = client.put("/api/project/project-id/info", data={
        "name": "project id",
        "authors": "asreview team",
        "description": "hello world"
    })
    assert response.status_code == 200


def test_get_project_info(client):
    """Test get info on the project"""

    response = client.get("/api/project/project-id/info")
    json_data = response.get_json()
    assert json_data["authors"] == "asreview team"
    assert json_data["dataset_path"] == "Hall_2012.csv"


def test_search_data(client):
    """Test search for papers"""

    response = client.get("/api/project/project-id/search?q=Software&n_max=10")
    json_data = response.get_json()

    assert "result" in json_data
    assert isinstance(json_data["result"], list)


def test_random_prior_papers(client):
    """Test get a selection of random papers to find exclusions"""

    response = client.get("/api/project/project-id/prior_random")
    json_data = response.get_json()

    assert "result" in json_data
    assert isinstance(json_data["result"], list)


def test_label_item(client):
    """Test label item"""

    response_irrelevant = client.post("/api/project/project-id/labelitem", data={
        "doc_id": 5509,
        "label": 0,
        "is_prior": 1
    })
    response_relevant = client.post("/api/project/project-id/labelitem", data={
        "doc_id": 58,
        "label": 1,
        "is_prior": 1
    })

    assert response_irrelevant.status_code == 200
    assert response_relevant.status_code == 200


def test_get_prior(client):
    """Test get all papers classified as prior documents"""

    response = client.get("/api/project/project-id/prior")
    json_data = response.get_json()

    assert "result" in json_data
    assert isinstance(json_data["result"], list)


def test_get_prior_stat(client):
    """Test get all papers classified as prior documents"""

    response = client.get("/api/project/project-id/prior_stats")
    json_data = response.get_json()

    assert "n_prior" in json_data
    assert isinstance(json_data, dict)


def test_list_algorithms(client):
    """Test get list of active learning models"""

    response = client.get("/api/algorithms")
    json_data = response.get_json()

    assert "classifier" in json_data.keys()
    assert "name" in json_data["classifier"][0].keys()
    assert isinstance(json_data, dict)


def test_set_algorithms(client):
    """Test set active learning model"""

    response = client.post("/api/project/project-id/algorithms", data={
        "model": "svm",
        "query_strategy": "random_max"
    })
    assert response.status_code == 200


def test_get_algorithms(client):
    """Test active learning model selection"""

    response = client.get("/api/project/project-id/algorithms")
    json_data = response.get_json()

    assert "model" in json_data
    assert "query_strategy" in json_data
    assert "svm" in json_data["model"]
    assert "random" in json_data["query_strategy"]
    assert isinstance(json_data, dict)


def test_start(client):
    """Test start training the model"""

    response = client.post("/api/project/project-id/start")
    assert response.status_code == 200


@pytest.mark.xfail(
    raises=KeyError,
    reason="status"
)
def test_init_model_ready(client):
    """Test check if trained model is available"""

    # wait the model ready
    time.sleep(8)

    response = client.get("/api/project/project-id/model/init_ready")
    json_data = response.get_json()

    assert json_data["status"] == 1


def test_clear_model_error(client):
    """Test clear model training error and retrain"""

    response_clear_error = client.delete("/api/project/project-id/model/clear_error")
    assert response_clear_error.status_code == 200

    # reset active learning model
    response_reset = client.post("/api/project/project-id/algorithms", data={
        "model": "svm",
        "query_strategy": "random"
    })
    assert response_reset.status_code == 200

    # retrain active learning model
    response_retrain = client.post("/api/project/project-id/start")
    assert response_retrain.status_code == 200

    # wait the model ready
    time.sleep(8)
    response_init_ready = client.get("/api/project/project-id/model/init_ready")
    json_data = response_init_ready.get_json()

    assert "status" in json_data
    assert json_data["status"] == 1


def test_export_result(client):
    """Test export result"""

    response_excel = client.get("/api/project/project-id/export?file_type=excel")
    response_csv = client.get("/api/project/project-id/export?file_type=csv")
    response_tsv = client.get("/api/project/project-id/export?file_type=tsv")

    assert response_excel.status_code == 200
    assert response_csv.status_code == 200
    assert response_tsv.status_code == 200


def test_export_project(client):
    """Test export the project file"""

    response = client.get("/api/project/project-id/export_project")
    assert response.status_code == 200


def test_finish_project(client):
    """Test mark a project as finished or not"""

    response = client.get("/api/project/project-id/finish")
    assert response.status_code == 200


def test_get_progress_info(client):
    """Test get progress info on the article"""

    response = client.get("/api/project/project-id/progress")
    json_data = response.get_json()
    assert isinstance(json_data, dict)


def test_get_progress_history(client):
    """Test get progress history on the article"""

    response = client.get("/api/project/project-id/progress_history")
    json_data = response.get_json()
    assert isinstance(json_data, list)


def test_get_progress_efficiency(client):
    """Test get cumulative number of inclusions by ASReview/at random"""

    response = client.get("/api/project/project-id/progress_efficiency")
    json_data = response.get_json()
    assert isinstance(json_data, list)


def test_classify_instance(client):
    """Test retrieve classification result"""

    response = client.post("/api/project/project-id/record/<doc_id>", data={
        "doc_id": 8208,
        "label": 1,
    })
    assert response.status_code == 200


def test_update_classify_instance(client):
    """Test update classification result"""

    response = client.put("/api/project/project-id/record/<doc_id>", data={
        "doc_id": 8208,
        "label": 0,
    })
    assert response.status_code == 200


def test_get_document(client):
    """Test retrieve documents in order of review"""

    response = client.get("/api/project/project-id/get_document")
    json_data = response.get_json()

    assert "result" in json_data
    assert isinstance(json_data, dict)


def test_delete_project(client):
    """Test get info on the article"""

    response = client.delete("/api/project/project-id/delete")
    assert response.status_code == 200



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.models.deprecated import _moved_warning

from asreview.models.feature_extraction.tfidf import Tfidf as _Tfidf
from asreview.models.feature_extraction.doc2vec import Doc2Vec as _Doc2Vec
from asreview.models.feature_extraction.embedding_idf \
    import EmbeddingIdf as _EmbeddingIdf
from asreview.models.feature_extraction.embedding_lstm \
    import EmbeddingLSTM as _EmbeddingLSTM
from asreview.models.feature_extraction.sbert import SBERT as _SBERT
from asreview.models.feature_extraction.utils \
    import get_feature_model as _get_feature_model
from asreview.models.feature_extraction.utils \
    import get_feature_class as _get_feature_class
from asreview.models.feature_extraction.utils \
    import list_feature_extraction as _list_feature_extraction

"""Deprecated, will be removed in version 1.0"""

Tfidf = _moved_warning(
    _Tfidf, "asreview.models.feature_extraction.Tfidf",
    "asreview.feature_extraction.Tfidf")
Doc2Vec = _moved_warning(
    _Doc2Vec, "asreview.models.feature_extraction.Doc2Vec",
    "asreview.feature_extraction.Doc2Vec")
EmbeddingIdf = _moved_warning(
    _EmbeddingIdf, "asreview.models.feature_extraction.EmbeddingIdf",
    "asreview.feature_extraction.EmbeddingIdf")
EmbeddingLSTM = _moved_warning(
    _EmbeddingLSTM, "asreview.models.feature_extraction.EmbeddingLSTM",
    "asreview.feature_extraction.EmbeddingLSTM")
SBERT = _moved_warning(
    _SBERT, "asreview.models.feature_extraction.SBERT",
    "asreview.feature_extraction.SBERT")
get_feature_model = _moved_warning(
    _get_feature_model, "asreview.models.feature_extraction.get_feature_model",
    "asreview.feature_extraction.get_feature_model")
get_feature_class = _moved_warning(
    _get_feature_class, "asreview.models.feature_extraction.get_feature_class",
    "asreview.feature_extraction.get_feature_class")
list_feature_extraction = _moved_warning(
    _list_feature_extraction,
    "asreview.models.feature_extraction.list_feature_extraction",
    "asreview.feature_extraction.list_feature_extraction")



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64encode, b64decode
from collections import OrderedDict
from datetime import datetime
from io import BytesIO

import numpy as np
from scipy.sparse.csr import csr_matrix
from scipy.sparse import save_npz, load_npz

from asreview.analysis.statistics import _get_labeled_order
from asreview.analysis.statistics import _get_last_proba_order
from asreview.state.base import BaseState
from asreview.settings import ASReviewSettings


def get_serial_list(array, dtype=None):
    if isinstance(array, np.ndarray):
        array = array.tolist()
    if dtype is not None:
        array = list(map(dtype, array))
    return array


class DictState(BaseState):
    """Class for storing the state of a review with no permanent storage."""
    version = "1.1"
    read_only = False

    def __init__(self, state_fp, *_, **__):
        super(DictState, self).__init__(state_fp)
        self.read_only = False

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def __str__(self):
        return self._print_state()

    def _print_state(self):
        self._state_dict["time"]["end_time"] = str(datetime.now())
        label_order, _ = _get_labeled_order(self)
        try:
            labels_assigned = self.get("labels")[label_order]
        except (KeyError, IndexError):
            return ""
        labels = list(zip(label_order, labels_assigned))

        state_str = "Labeled during review:\n\n"
        for label in labels:
            state_str += f"{label[0]} => {label[1]}\n"

        pool_order = _get_last_proba_order(self)
        if len(pool_order) > 0:
            state_str += "\n\n Most likely included according to ASReview:\n\n"
            for idx in pool_order:
                state_str += f"{idx}\n"

        return state_str

    def _add_to_state(self, new_dict, i, append_result=False):
        # Find the first number that has not yet been used.
        results = self._state_dict["results"]
        if i is None:
            if (len(results) > 0 and
                    set(new_dict.keys()).isdisjoint(results[-1])):
                i = len(results) - 1
            else:
                i = len(results)

        while i >= len(results):
            results.append({})

        for key in new_dict:
            if key in results[i] and append_result:
                results[i][key].extend(new_dict[key])
            else:
                results[i][key] = new_dict[key]

    def _add_as_data(self, as_data, feature_matrix=None):
        record_table = as_data.record_ids
        data_hash = as_data.hash()

        if "data_properties" not in self._state_dict:
            self._state_dict["data_properties"] = {}

        data_properties = self._state_dict["data_properties"]
        if data_hash not in data_properties:
            data_properties[data_hash] = {}
        data_properties[data_hash]["record_table"] = record_table.tolist()
        if feature_matrix is None:
            return
        if isinstance(feature_matrix, np.ndarray):
            encoded_X = feature_matrix.tolist()
            matrix_type = "ndarray"
        elif isinstance(feature_matrix, csr_matrix):
            with BytesIO() as f:
                save_npz(f, feature_matrix)
                f.seek(0)
                encoded_X = b64encode(f.read()).decode('ascii')
            matrix_type = "csr_matrix"
        else:
            encoded_X = feature_matrix
            matrix_type = "unknown"
        data_properties[data_hash]["feature_matrix"] = encoded_X
        data_properties[data_hash]["matrix_type"] = matrix_type

    def get_feature_matrix(self, data_hash):
        my_data = self._state_dict["data_properties"][data_hash]
        encoded_X = my_data["feature_matrix"]
        matrix_type = my_data["matrix_type"]
        if matrix_type == "ndarray":
            return np.array(encoded_X)
        elif matrix_type == "csr_matrix":
            with BytesIO(b64decode(encoded_X)) as f:
                return load_npz(f)
        return encoded_X

    def get_current_queries(self):
        str_queries = self._state_dict["current_queries"]
        return {int(key): value for key, value in str_queries.items()}

    def set_current_queries(self, current_queries):
        self._state_dict["current_queries"] = current_queries

    def is_empty(self):
        return len(self._state_dict["results"]) == 0

    def set_labels(self, y):
        self._state_dict["labels"] = y.tolist()

    def set_final_labels(self, y):
        self._state_dict["final_labels"] = y.tolist()

    @property
    def settings(self):
        settings = self._state_dict.get("settings", None)
        if settings is None:
            return None
        return ASReviewSettings(**settings)

    @settings.setter
    def settings(self, settings):
        self._state_dict["settings"] = vars(settings)

    def add_classification(self, idx, labels, methods, query_i):
        # Ensure that variables are serializable
        idx = get_serial_list(idx, int)
        labels = get_serial_list(labels, int)
        methods = get_serial_list(methods, str)

        new_dict = {'labelled': list(zip(idx, labels, methods))}
        self._add_to_state(new_dict, query_i, append_result=True)

    def add_proba(self, pool_idx, train_idx, proba, query_i):
        new_dict = {
            "pool_idx": get_serial_list(pool_idx, int),
            "train_idx": get_serial_list(train_idx, int),
            "proba": get_serial_list(proba, float),
        }
        self._add_to_state(new_dict, query_i)

    def n_queries(self):
        return len(self._state_dict["results"])

    def get(self, variable, query_i=None, idx=None):
        if query_i is not None:
            res = self._state_dict["results"][query_i]

        array = None
        label_vars = ["label_idx", "inclusions", "label_methods"]
        if variable in label_vars:
            array_id = label_vars.index(variable)
            if variable == "label_methods":
                dtype = str
            else:
                dtype = np.int
            array = np.array([x[array_id] for x in res["labelled"]],
                             dtype=dtype)
        elif variable == "labels":
            array = np.array(self._state_dict["labels"], dtype=np.int)
        elif variable == "final_labels":
            array = np.array(self._state_dict["final_labels"], dtype=np.int)
        elif variable == "proba":
            array = np.array(res["proba"], dtype=np.float)
        elif variable == "train_idx":
            array = np.array(res["train_idx"], dtype=np.int)
        elif variable == "pool_idx":
            array = np.array(res["pool_idx"], dtype=np.int)
        if array is None:
            return None
        if idx is not None:
            return array[idx]
        return array

    def delete_last_query(self):
        self._state_dict["results"].pop()

    def initialize_structure(self):
        from asreview import __version__ as asr_version
        self._state_dict = OrderedDict({
            "time": {
                "start_time": str(datetime.now())
            },
            "version": self.version,
            "software_version": asr_version,
            "settings": {},
            "results": [],
        })

    def close(self):
        if not self.read_only:
            self.save()

    def save(self):
        print(self)

    def restore(self, *_, **__):
        self.initialize_structure()



# Copyright 2019-2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Uncertainty sampling while saving probabilities."""

import numpy as np

from asreview.models.query.base import ProbaQueryStrategy


class UncertaintyQuery(ProbaQueryStrategy):
    """Uncertainty query strategy.

    Choose the most uncertain samples according to the model (i.e. closest to
    0.5 probability). Doesn’t work very well in the case of LSTM’s, since the
    probabilities are rather arbitrary.

    """

    name = "uncertainty"
    label = "Uncertainty"

    def _query(self, X, pool_idx, n_instances=1, proba=None):
        uncertainty = 1 - np.max(proba[pool_idx], axis=1)
        query_idx = np.argsort(-uncertainty)[:n_instances]
        return pool_idx[query_idx], X[pool_idx[query_idx]]



# Copyright 2019-2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket


def check_port_in_use(host, port):
    """Check if port is already in use.

    Arguments
    ---------
    host: str
        The current host.
    port: int
        The host port to be checked.

    Returns
    -------
    bool:
        True if port is in use, false otherwise.
    """
    logging.info(
        f"Checking if host and port are available :: {host}:{port}"
    )
    host = host.replace('https://', '').replace('http://', '')
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0



# Copyright 2019-2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    from sentence_transformers.SentenceTransformer import SentenceTransformer  # noqa
except ImportError:
    ST_AVAILABLE = False
else:
    ST_AVAILABLE = True

from asreview.models.feature_extraction.base import BaseFeatureExtraction


def _check_st():
    if not ST_AVAILABLE:
        raise ImportError(
            "Install sentence-transformers package"
            " to use Sentence BERT.")


class SBERT(BaseFeatureExtraction):
    """Sentence BERT feature extraction technique.

    Feature extraction technique based on Sentence BERT. Implementation based on
    the `sentence_transformers <https://github.com/UKPLab/sentence-
    transformers>`__ package. It is relatively slow.

    .. note::

        This feature extraction technique requires ``sentence_transformers``
        to be installed. Use ``pip install sentence_transformers`` or install
        all optional ASReview dependencies with ``pip install asreview[all]``

    """

    name = "sbert"
    label = "Sentence BERT"

    def transform(self, texts):

        _check_st()

        model = SentenceTransformer('bert-base-nli-mean-tokens')
        X = np.array(model.encode(texts))
        return X



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.utils import list_model_names
from asreview.utils import _model_class_from_entry_point


def list_feature_extraction():
    """List available feature extraction method classes.

    Returns
    -------
    list:
        Classes of available feature extraction methods in alphabetical order.
    """
    model_class = [
        get_feature_class(name)
        for name in list_model_names(entry_name="asreview.models.feature_extraction")
    ]

    return model_class


def get_feature_class(name):
    """Get class of feature extraction from string.

    Arguments
    ---------
    name: str
        Name of the feature model, e.g. 'doc2vec', 'tfidf' or 'embedding-lstm'.

    Returns
    -------
    BaseFeatureExtraction:
        Class corresponding to the name.
    """
    return _model_class_from_entry_point(name, "asreview.models.feature_extraction")


def get_feature_model(name, *args, random_state=None, **kwargs):
    """Get an instance of a feature extraction model from a string.

    Arguments
    ---------
    name: str
        Name of the feature extraction model.
    *args:
        Arguments for the feature extraction model.
    **kwargs:
        Keyword arguments for thefeature extraction  model.

    Returns
    -------
    BaseFeatureExtraction:
        Initialized instance of feature extraction algorithm.
    """
    model_class = get_feature_class(name)
    try:
        return model_class(*args, random_state=random_state, **kwargs)
    except TypeError:
        return model_class(*args, **kwargs)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os
from pathlib import Path

import numpy as np
from scipy import stats

from asreview.state.utils import states_from_dir, state_from_file
from asreview.analysis.statistics import _get_labeled_order
from asreview.analysis.statistics import _get_limits
from asreview.analysis.statistics import _find_inclusions
from asreview.analysis.statistics import _get_last_proba_order


class Analysis():
    """Analysis object to do statistical analysis on state files."""
    def __init__(self, states, key=None):
        """Class to analyse state files.

        Arguments
        ---------
        states: list, BaseLogger
            Either a list of states (opened files) or a single state.
        key: str
            Give a name to the analysis.
        """
        if isinstance(states, list):
            states = {i: state for i, state in enumerate(states)}

        # Sometimes an extra dataset is present in the state_file(s).
        # These signify not the labels on which the model was trained, but the
        # ones that were included in the end (or some other intermediate step.
        self.final_labels = None
        self.labels = None
        self.empty = True

        if states is None:
            return

        self.key = key
        self.states = states
        self.num_runs = len(self.states)
        if self.num_runs == 0:
            return

        self._first_file = list(self.states.keys())[0]
        self.labels = self.states[self._first_file].get('labels')
        try:
            self.final_labels = self.states[self._first_file].get(
                'final_labels')
        except KeyError:
            pass
        self.empty = False
        self.inc_found = {}

    @classmethod
    def from_dir(cls, data_dir, prefix="", key=None):
        """Create an Analysis object from a directory.

        Arguments
        ---------
        data_dir: str
            Directory to read the state files from.
        prefix: str
            Only assume files starting with this prefix are state files.
            Ignore all other files.
        key: str
            Name for the analysis object.
        """
        if key is None:
            key = os.path.basename(os.path.normpath(data_dir))
        states = states_from_dir(data_dir, prefix=prefix)
        analysis_inst = cls(states, key=key)
        if analysis_inst.empty:
            return None

        analysis_inst.data_dir = data_dir
        return analysis_inst

    @classmethod
    def from_file(cls, data_fp, key=None):
        """Create an Analysis object from a file.

        Arguments
        ---------
        data_fp: str
            Path to state file to analyse.
        key: str
            Name for analysis object.
        """
        if key is None:
            key = os.path.basename(os.path.normpath(data_fp))
        state = state_from_file(data_fp)
        analysis_inst = cls(state, key=key)
        if analysis_inst.empty:
            return None

        analysis_inst.data_path = data_fp
        return analysis_inst

    @classmethod
    def from_path(cls, data_path, prefix="", key=None):
        """Create an Analysis object from either a file or a directory."""
        if Path(data_path).is_file():
            return cls.from_file(data_path, key=key)
        return cls.from_dir(data_path, prefix, key=key)

    def inclusions_found(self,
                         result_format="fraction",
                         final_labels=False,
                         **kwargs):
        """Get the number of inclusions at each point in time.

        Caching is used to prevent multiple calls being expensive.

        Arguments
        ---------
        result_format: str
            The format % or # of the returned values.
        final_labels: bool
            If true, use the final_labels instead of labels for analysis.

        Returns
        -------
        tuple:
            Three numpy arrays with x, y, error_bar.
        """
        if final_labels:
            labels = self.final_labels
        else:
            labels = self.labels

        fl = final_labels
        if fl not in self.inc_found:
            # Compute the comclusions if not found in cache.
            self.inc_found[fl] = {}
            avg, err, iai, ninit = self._get_inc_found(labels=labels, **kwargs)
            self.inc_found[fl]["avg"] = avg
            self.inc_found[fl]["err"] = err
            self.inc_found[fl]["inc_after_init"] = iai
            self.inc_found[fl]["n_initial"] = ninit
        dx = 0
        dy = 0
        x_norm = len(labels) - self.inc_found[fl]["n_initial"]
        y_norm = self.inc_found[fl]["inc_after_init"]

        if result_format == "percentage":
            x_norm /= 100
            y_norm /= 100
        elif result_format == "number":
            x_norm /= len(labels)
            y_norm /= self.inc_found[fl]["inc_after_init"]

        # extend the arrays if all relevant records have been found
        total_inc_found = int(self.inc_found[fl]["avg"][-1])
        if total_inc_found == self.inc_found[fl]["inc_after_init"]:
            norm_xr = np.arange(1, 1 + len(labels) -
                                self.inc_found[fl]["n_initial"]) / x_norm
            missing_y = len(labels) - self.inc_found[fl]["n_initial"] - \
                len(self.inc_found[fl]["avg"])
            self.inc_found[fl]["avg"].extend([total_inc_found] * missing_y)
        else:
            norm_xr = (np.arange(1, len(self.inc_found[fl]["avg"]) + 1) - dx) \
                / x_norm

        norm_yr = (np.array(self.inc_found[fl]["avg"]) - dy) / y_norm
        norm_y_err = np.array(self.inc_found[fl]["err"]) / y_norm

        return norm_xr, norm_yr, norm_y_err

    def _get_inc_found(self, labels=False):
        """Get the number of inclusions (without formatting)."""
        inclusions_found = []

        for state in self.states.values():
            inclusions, inc_after_init, n_initial = _find_inclusions(
                state, labels)
            inclusions_found.append(inclusions)

        inc_found_avg = []
        inc_found_err = []
        for i_instance in itertools.count():
            cur_vals = []
            for i_file in range(self.num_runs):
                try:
                    cur_vals.append(inclusions_found[i_file][i_instance])
                except IndexError:
                    pass
            if len(cur_vals) == 0:
                break
            if len(cur_vals) == 1:
                err = cur_vals[0]
            else:
                err = stats.sem(cur_vals)
            avg = np.mean(cur_vals)
            inc_found_avg.append(avg)
            inc_found_err.append(err)

        if self.num_runs == 1:
            inc_found_err = np.zeros(len(inc_found_err))

        return inc_found_avg, inc_found_err, inc_after_init, n_initial

    def wss(self, val=100, x_format="percentage", **kwargs):
        """Get the WSS (Work Saved Sampled) value.

        Arguments
        ---------
        val:
            At which recall, between 0 and 100.
        x_format:
            Format for position of WSS value in graph.

        Returns
        -------
        tuple:
            Tuple consisting of WSS value, x_positions, y_positions of WSS bar.
        """
        norm_xr, norm_yr, _ = self.inclusions_found(result_format="percentage",
                                                    **kwargs)

        if x_format == "number":
            x_return, y_result, _ = self.inclusions_found(
                result_format="number", **kwargs)
            y_max = self.inc_found[False]["inc_after_init"]
            y_coef = y_max / (len(self.labels) -
                              self.inc_found[False]["n_initial"])
        else:
            x_return = norm_xr
            y_result = norm_yr
            y_max = 1.0
            y_coef = 1.0

        for i in range(len(norm_yr)):
            if norm_yr[i] >= val - 1e-6:
                return (norm_yr[i] - norm_xr[i], (x_return[i], x_return[i]),
                        (x_return[i] * y_coef, y_result[i]))
        return (None, None, None)

    def rrf(self, val=10, x_format="percentage", **kwargs):
        """Get the RRF (Relevant References Found).

        Arguments
        ---------
        val:
            At which recall, between 0 and 100.
        x_format:
            Format for position of RRF value in graph.

        Returns
        -------
        tuple:
            Tuple consisting of RRF value, x_positions, y_positions of RRF bar.

        """
        norm_xr, norm_yr, _ = self.inclusions_found(result_format="percentage",
                                                    **kwargs)

        if x_format == "number":
            x_return, y_return, _ = self.inclusions_found(
                result_format="number", **kwargs)
        else:
            x_return = norm_xr
            y_return = norm_yr

        for i in range(len(norm_yr)):
            if norm_xr[i] >= val - 1e-6:
                return (norm_yr[i], (x_return[i], x_return[i]), (0,
                                                                 y_return[i]))
        return (None, None, None)

    def avg_time_to_discovery(self, result_format="number"):
        """Estimate the Time to Discovery (TD) for each paper.

        Get the best/last estimate on how long it takes to find a paper.

        Arguments
        ---------
        result_format: str
            Desired output format: "number", "fraction" or "percentage".

        Returns
        -------
        dict:
            For each inclusion, key=paper_id, value=avg time.
        """
        labels = self.labels
        one_labels = np.where(labels == 1)[0]
        time_results = {label: [] for label in one_labels}

        # Iterate over all state files
        for state in self.states.values():
            # Get the order in which records were labeled
            label_order, n = _get_labeled_order(state)
            # Get the ranking of all papers at the last query
            proba_order = _get_last_proba_order(state)

            # Adjust factor, depending on the desired output format
            if result_format == "percentage":
                time_mult = 100 / (len(labels) - n)
            elif result_format == "fraction":
                time_mult = 1 / (len(labels) - n)
            else:
                time_mult = 1

            # Get the time to discovery
            for i_time, idx in enumerate(label_order[n:]):
                # for all inclusions that were found/labeled
                if labels[idx] == 1:
                    time_results[idx].append(time_mult * (i_time + 1))
            for i_time, idx in enumerate(proba_order):
                # for all inclusions that weren't found/labeled
                if labels[idx] == 1 and idx not in label_order[:n]:
                    time_results[idx].append(time_mult *
                                             (i_time + len(label_order) + 1))

        results = {}

        # Merge the results of all state files
        for label, trained_time in time_results.items():
            if len(trained_time) > 0:
                results[label] = np.average(trained_time)

        return results

    def limits(self, prob_allow_miss=[0.1], result_format="percentage"):
        """For each query, compute the number of papers for a criterium.

        A criterium is the average number of papers missed. For example,
        with 0.1, the criterium is that after reading x papers, there is
        (about) a 10% chance that one paper is not included. Another example,
        with 2.0, there are on average 2 papers missed after reading x papers.
        The value for x is returned for each query and probability by the
        function.

        Arguments
        ---------
        prob_allow_miss: list, float
            Sets the criterium for how many papers can be missed.

        returns
        -------
        dict:
            One entry, "x_range" with the number of papers read.
            List, "limits" of results for each probability and
            at # papers read.
        """
        if not isinstance(prob_allow_miss, list):
            prob_allow_miss = [prob_allow_miss]
        state = self.states[self._first_file]
        n_queries = state.n_queries()
        results = {
            "x_range": [],
            "limits": [[] for _ in range(len(prob_allow_miss))],
        }

        n_train = 0
        _, n_initial = _get_labeled_order(state)
        for query_i in range(n_queries):
            new_limits = _get_limits(self.states,
                                     query_i,
                                     self.labels,
                                     proba_allow_miss=prob_allow_miss)

            try:
                new_train_idx = state.get("train_idx", query_i)
            except KeyError:
                new_train_idx = None

            if new_train_idx is not None:
                n_train = len(new_train_idx)

            if new_limits is not None:
                if result_format == "percentage":
                    normalizer = 100 / (len(self.labels) - n_initial)
                else:
                    normalizer = 1
                results["x_range"].append((n_train - n_initial) * normalizer)
                for i_prob in range(len(prob_allow_miss)):
                    results["limits"][i_prob].append(
                        (new_limits[i_prob] - n_initial) * normalizer)

        if result_format == "percentage":
            res_dtype = np.float
        else:
            res_dtype = np.int

        results["x_range"] = np.array(results["x_range"], dtype=res_dtype)
        for i_prob in range(len(prob_allow_miss)):
            results["limits"][i_prob] = np.array(results["limits"][i_prob],
                                                 res_dtype)
        return results

    def close(self):
        """Close states."""
        for state in self.states.values():
            state.close()



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# based on https://github.com/pypa/sampleproject - MIT License

# Always prefer setuptools over distutils
import re
import subprocess
from io import open
from os import path
from setuptools import setup
from setuptools import find_packages
from setuptools import Command

import versioneer


def get_long_description():
    """Get project description based on README."""
    here = path.abspath(path.dirname(__file__))

    # Get the long description from the README file
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

    # remove emoji
    long_description = re.sub(r"\:[a-z_]+\:", "", long_description)

    return long_description


DEPS = {
    "sbert": ['sentence_transformers'],
    "doc2vec": ['gensim'],
    "tensorflow": ['tensorflow~=2.0'],
    "dev": ['check-manifest'],
    'test': ['coverage', 'pytest'],
}
DEPS['all'] = DEPS['sbert'] + DEPS['doc2vec'] + DEPS['dev']
DEPS['all'] += DEPS['tensorflow']


class CompileAssets(Command):
    """
    Compile and build the frontend assets using npm and webpack.

    Registered as cmdclass in setup() so it can be called with
    ``python setup.py compile_assets``.
    """

    description = "Compile and build the frontend assets"
    user_options = []

    def initialize_options(self):
        """Set default values for options."""

    def finalize_options(self):
        """Set final values for options."""

    def run(self):
        """Run a command to compile and build assets."""
        subprocess.check_call('sh ./asreview/webapp/compile_assets.sh',
                              shell=True)


def get_cmdclass():
    cmdclass = versioneer.get_cmdclass()
    cmdclass["compile_assets"] = CompileAssets
    return cmdclass


setup(
    name='asreview',
    version=versioneer.get_version(),
    cmdclass=get_cmdclass(),
    description='Active learning for Systematic Reviews',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/asreview/asreview',
    author='ASReview Core Development Team',
    author_email='asreview@uu.nl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Framework :: Flask',
    ],
    keywords='systematic review machine-learning',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    package_data={'asreview': [
        'webapp/build/*',
        'webapp/build/static/*/*',
    ]},
    python_requires='~=3.7',
    install_requires=[
        'numpy',
        'sklearn',
        'pandas',
        'rispy~=0.6.0',
        'dill',
        'h5py',
        'xlrd>=1.0.0',
        'setuptools',
        'flask>=2.0',
        'flask_cors',
        'openpyxl',
        'gevent',
    ],
    extras_require=DEPS,
    entry_points={
        'console_scripts': [
            'asreview=asreview.__main__:main',
        ],
        'asreview.entry_points': [
            'lab=asreview.entry_points:LABEntryPoint',
            'oracle=asreview.entry_points:OracleEntryPoint',  # deprecated (use lab)
            'web_run_model = asreview.entry_points:WebRunModelEntryPoint',
            'simulate=asreview.entry_points:SimulateEntryPoint',
            'simulate-batch = asreview.entry_points:BatchEntryPoint',
            'algorithms = asreview.entry_points:AlgorithmsEntryPoint',
        ],
        'asreview.readers': [
            '.csv = asreview.io.csv_reader:read_csv',
            '.tab = asreview.io.csv_reader:read_csv',
            '.tsv = asreview.io.csv_reader:read_csv',
            '.ris = asreview.io.ris_reader:read_ris',
            '.txt = asreview.io.ris_reader:read_ris',
            '.xlsx = asreview.io.excel_reader:read_excel',
            '.xml = asreview.io.pubmed_xml_reader: read_pubmed_xml',
        ],
        'asreview.datasets': [
            'benchmark = asreview.datasets:BenchmarkDataGroup',
        ],
        'asreview.models.classifiers': [
            'svm = asreview.models.classifiers:SVMClassifier',
            'nb = asreview.models.classifiers:NaiveBayesClassifier',
            'rf = asreview.models.classifiers:RandomForestClassifier',
            'nn-2-layer = asreview.models.classifiers:NN2LayerClassifier',
            'logistic = asreview.models.classifiers:LogisticClassifier',
            'lstm-base = asreview.models.classifiers:LSTMBaseClassifier',
            'lstm-pool = asreview.models.classifiers:LSTMPoolClassifier',
        ],
        'asreview.models.feature_extraction': [
            'doc2vec = asreview.models.feature_extraction:Doc2Vec',
            'embedding-idf = asreview.models.feature_extraction:EmbeddingIdf',
            'embedding-lstm = asreview.models.feature_extraction:EmbeddingLSTM',
            'sbert = asreview.models.feature_extraction:SBERT',
            'tfidf = asreview.models.feature_extraction:Tfidf',
        ],
        'asreview.models.balance': [
            "simple = asreview.models.balance:SimpleBalance",
            "double = asreview.models.balance:DoubleBalance",
            "triple = asreview.models.balance:TripleBalance",
            "undersample = asreview.models.balance:UndersampleBalance",
        ],
        'asreview.models.query': [
            "max = asreview.models.query.max:MaxQuery",
            "random = asreview.models.query.random:RandomQuery",
            "uncertainty = asreview.models.query.uncertainty:UncertaintyQuery",
            "cluster = asreview.models.query.cluster:ClusterQuery",
            "max_random = asreview.models.query.mixed:MaxRandomQuery",
            "max_uncertainty = asreview.models.query.mixed:MaxUncertaintyQuery",
        ]
    },
    project_urls={
        'Bug Reports': 'https://github.com/asreview/asreview/issues',
        'Source': 'https://github.com/asreview/asreview/',
    },
)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

import numpy as np
from sklearn.exceptions import NotFittedError

from asreview.models.base import BaseModel


class BaseQueryStrategy(BaseModel):
    """Abstract class for query strategies."""

    name = "base-query"

    @abstractmethod
    def query(self,
              X,
              classifier=None,
              pool_idx=None,
              n_instances=1,
              shared={}):
        """Query new instances.

        Arguments
        ---------
        X: numpy.ndarray
            Feature matrix to choose samples from.
        classifier: SKLearnModel
            Trained classifier to compute probabilities if they are necessary.
        pool_idx: numpy.ndarray
            Indices of samples that are still in the pool.
        n_instances: int
            Number of instances to query.
        shared: dict
            Dictionary for exchange between query strategies and others.
            It is mainly used to store the current class probabilities,
            and the source of the queries; which query strategy has produced
            which index.
        """
        raise NotImplementedError


class ProbaQueryStrategy(BaseQueryStrategy):
    name = "proba"

    def query(self, X, classifier, pool_idx=None, n_instances=1, shared={}):
        """Query method for strategies which use class probabilities.
        """
        n_samples = X.shape[0]
        if pool_idx is None:
            pool_idx = np.arange(n_samples)

        proba = shared.get('pred_proba', [])
        if len(proba) != n_samples:
            try:
                proba = classifier.predict_proba(X)
            except NotFittedError:
                proba = np.ones(shape=(n_samples, ))
            shared['pred_proba'] = proba
        query_idx, X_query = self._query(X, pool_idx, n_instances, proba)

        for idx in query_idx:
            shared['current_queries'][idx] = self.name

        return query_idx, X_query

    @abstractmethod
    def _query(self, X, pool_idx, n_instances, proba):
        raise NotImplementedError


class NotProbaQueryStrategy(BaseQueryStrategy):
    name = "not_proba"

    def query(self, X, classifier, pool_idx=None, n_instances=1, shared={}):
        """Query method for strategies which do not use class probabilities
        """
        n_samples = X.shape[0]
        if pool_idx is None:
            pool_idx = np.arange(n_samples)

        query_idx, X_query = self._query(X, pool_idx, n_instances)

        for idx in query_idx:
            shared['current_queries'][idx] = self.name

        return query_idx, X_query

    @abstractmethod
    def _query(self, X, pool_idx, n_instances, proba):
        raise NotImplementedError



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.models.deprecated import _moved_warning

from asreview.models.balance import SimpleBalance as _SimpleBalance
from asreview.models.balance import DoubleBalance as _DoubleBalance
from asreview.models.balance import TripleBalance as _TripleBalance
from asreview.models.balance import UndersampleBalance as _UndersampleBalance
from asreview.models.balance import get_balance_model as _get_balance_model
from asreview.models.balance import get_balance_class as _get_balance_class
from asreview.models.balance import list_balance_strategies as _list_balance_strategies

"""Deprecated, will be removed in version 1.0"""

SimpleBalance = _moved_warning(
    _SimpleBalance, "asreview.models.balance.SimpleBalance",
    "asreview.balance_strategies.SimpleBalance")
DoubleBalance = _moved_warning(
    _DoubleBalance, "asreview.models.balance.DoubleBalance",
    "asreview.balance_strategies.DoubleBalance")
TripleBalance = _moved_warning(
    _TripleBalance, "asreview.models.balance.TripleBalance",
    "asreview.balance_strategies.TripleBalance")
UndersampleBalance = _moved_warning(
    _UndersampleBalance, "asreview.models.balance.UndersampleBalance",
    "asreview.balance_strategies.UndersampleBalance")
get_balance_model = _moved_warning(
    _get_balance_model, "asreview.models.balance.get_balance_model",
    "asreview.balance_strategies.get_balance_model")
get_balance_class = _moved_warning(
    _get_balance_class, "asreview.models.balance.get_balance_class",
    "asreview.balance_strategies.get_balance_class")
list_balance_strategies = _moved_warning(
    _list_balance_strategies,
    "asreview.models.balance.list_balance_strategies",
    "asreview.balance_strategies.list_balance_strategies")



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

AVAILABLE_CLI_MODI = ["simulate"]
AVAILABLE_REVIEW_CLASSES = ["simulate", "minimal"]

LABEL_NA = -1

KERAS_MODELS = ["lstm_base", "lstm_pool"]

# CLI defaults
DEFAULT_MODEL = "nb"
DEFAULT_QUERY_STRATEGY = "max"
DEFAULT_BALANCE_STRATEGY = "double"
DEFAULT_FEATURE_EXTRACTION = "tfidf"
DEFAULT_N_INSTANCES = 1
DEFAULT_N_PRIOR_INCLUDED = 1
DEFAULT_N_PRIOR_EXCLUDED = 1

GITHUB_PAGE = "https://github.com/asreview/asreview"
EMAIL_ADDRESS = "asreview@uu.nl"

STATE_EXTENSIONS = [".h5", ".hdf5", ".he5", ".json"]
LOGGER_EXTENSIONS = STATE_EXTENSIONS

COLUMN_DEFINITIONS = {
    # included
    "included": [
        "final_included", "label", "label_included", "included_label",
        "included_final", "included", "included_flag", "include"
    ],
    # abstract included (pending deprecation)
    "abstract_included": [
        "abstract_included", "included_abstract", "included_after_abstract",
        "label_abstract_screening"
    ],
    "title": ["title", "primary_title"],
    "authors": ["authors", "author names", "first_authors"],
    "abstract": ["abstract", "abstract note", "notes_abstract"],
    "keywords": ["keywords"],
    "doi": ["doi"],
}



# Copyright 2019-2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from math import floor

import numpy as np
from scipy.sparse import issparse
from scipy.sparse import vstack

from asreview.models.query.base import BaseQueryStrategy
from asreview.models.query.utils import get_query_model
from asreview.utils import get_random_state


def interleave(n_samples, n_strat_1, random_state):
    """Interleave the order of the samples of two different strategies.

    While the decisions of which indices to sample are made one after
    the other, it is nicer if the order of the actual samples is mixed up.
    Instead of mixing it the easy way, I decided it should be as nice as
    possible.

    Parameters
    ----------
    n_samples: int
        Total number of samples to mix.
    n_strat_1: int
        Number of samples of the first strategy.
    random_state: int, numpy.RandomState
        RNG.

    Returns
    -------
    numpy.array:
        Order of samples, [0, n_samples).
    """
    n_strat_2 = n_samples - n_strat_1

    # Determine which of the strategies has more samples.
    if n_strat_1 >= n_strat_2:
        max_idx = np.arange(n_strat_1)
        min_idx = n_strat_1 + np.arange(n_strat_2)
    else:
        max_idx = n_strat_1 + np.arange(n_strat_2)
        min_idx = np.arange(n_strat_1)

    # Insert the strategy with less samples at these positions.
    insert_positions = np.sort(
        random_state.choice(
            np.arange(len(max_idx)), len(min_idx), replace=False))

    # Actually do the inserts.
    new_positions = np.zeros(n_samples, dtype=int)
    i_strat_min = 0
    for i_strat_max in range(len(max_idx)):
        new_positions[i_strat_max + i_strat_min] = max_idx[i_strat_max]
        if (i_strat_min < len(min_idx) and
                insert_positions[i_strat_min] == i_strat_max):
            new_positions[i_strat_min + i_strat_max + 1] = min_idx[i_strat_min]
            i_strat_min += 1
    return new_positions


class MixedQuery(BaseQueryStrategy):
    """Mixed query strategy.

    The idea is to use two different query strategies at the same time with a
    ratio of one to the other. A mix of two query strategies is used. For
    example mixing max and random sampling with a mix ratio of 0.95 would mean
    that at each query 95% of the instances would be sampled with the max
    query strategy after which the remaining 5% would be sampled with the
    random query strategy. It would be called the `max_random` query strategy.
    Every combination of primitive query strategy is possible.

    Arguments
    ---------
    strategy_1: str
        Name of the first query strategy.
    strategy_2: str
        Name of the second query strategy.
    mix_ratio: float
        Portion of queries done by the first strategy. So a mix_ratio of
        0.95 means that 95% of the time query strategy 1 is used and 5% of
        the time query strategy 2.
    **kwargs: dict
        Keyword arguments for the two strategy. To specify which of the
        strategies the argument is for, prepend with the name of the query
        strategy and an underscore, e.g. 'max' for maximal sampling.
    """

    def __init__(self,
                 strategy_1="max",
                 strategy_2="random",
                 mix_ratio=0.95,
                 random_state=None,
                 **kwargs):
        """Initialize the Mixed query strategy."""
        super(MixedQuery, self).__init__()
        kwargs_1 = {}
        kwargs_2 = {}
        for key, value in kwargs.items():
            if key.startswith(strategy_1):
                new_key = key[len(strategy_1) + 1:]
                kwargs_1[new_key] = value
            elif key.starts_with(strategy_2):
                new_key = key[len(strategy_2) + 1:]
                kwargs_2[new_key] = value
            else:
                logging.warn(f"Key {key} is being ignored for the mixed "
                             "({strategy_1}, {strategy_2}) query strategy.")

        self.strategy_1 = strategy_1
        self.strategy_2 = strategy_2

        self.query_model1 = get_query_model(strategy_1, **kwargs_1)
        self.query_model2 = get_query_model(strategy_2, **kwargs_2)

        self._random_state = get_random_state(random_state)
        if "random_state" in self.query_model1.default_param:
            self.query_model1 = get_query_model(
                strategy_1, **kwargs_1, random_state=self._random_state)
        if "random_state" in self.query_model2.default_param:
            self.query_model2 = get_query_model(
                strategy_2, **kwargs_2, random_state=self._random_state)
        self.mix_ratio = mix_ratio

    def query(self, X, classifier, pool_idx=None, n_instances=1, shared={}):
        n_samples = X.shape[0]
        if pool_idx is None:
            pool_idx = np.arange(n_samples)

        # Split the number of instances for the query strategies.
        n_instances_1 = floor(n_instances * self.mix_ratio)
        leftovers = n_instances * self.mix_ratio - n_instances_1
        if self._random_state.random_sample() < leftovers:
            n_instances_1 += 1
        n_instances_2 = n_instances - n_instances_1

        # Perform the query with strategy 1.
        query_idx_1, X_1 = self.query_model1.query(
            X,
            classifier,
            pool_idx=pool_idx,
            n_instances=n_instances_1,
            shared=shared)

        # Remove the query indices from the pool.
        train_idx = np.delete(np.arange(n_samples), pool_idx, axis=0)
        train_idx = np.append(train_idx, query_idx_1)
        new_pool_idx = np.delete(np.arange(n_samples), train_idx, axis=0)

        # Perform the query with strategy 2.
        query_idx_2, X_2 = self.query_model2.query(
            X,
            classifier,
            pool_idx=new_pool_idx,
            n_instances=n_instances_2,
            shared=shared)

        query_idx = np.append(query_idx_1, query_idx_2)

        if n_instances_1 == 0:
            X = X_2
        elif n_instances_2 == 0:
            X = X_1
        else:
            if issparse(X_1) and issparse(X_2):
                X = vstack([X_1, X_2]).tocsr()
            else:
                X = np.concatenate((X_1, X_2), axis=0)

        # Remix the two strategies without changing the order within.
        new_order = interleave(
            len(query_idx), len(query_idx_1), self._random_state)
        return query_idx[new_order], X[new_order]

    def full_hyper_space(self):
        from hyperopt import hp

        space_1, choices_1 = self.query_model1.hyper_space()
        space_2, choices_2 = self.query_model2.hyper_space()
        parameter_space = {}
        hyper_choices = {}
        for key, value in space_1.items():
            new_key = "qry_" + self.strategy_1 + key[4:]
            parameter_space[new_key] = value
            hyper_choices[new_key] = choices_1[key]

        for key, value in space_2.items():
            new_key = "qry_" + self.strategy_2 + key[4:]
            parameter_space[new_key] = value
            hyper_choices[new_key] = choices_2[key]

        parameter_space["qry_mix_ratio"] = hp.uniform("qry_mix_ratio", 0, 1)

        return parameter_space, hyper_choices

    @property
    def name(self):
        return "_".join([self.strategy_1, self.strategy_2])


class MaxRandomQuery(MixedQuery):
    """Mixed (95% Maximum and 5% Random) query strategy.

    A mix of maximum and random query strategies with a mix ratio of 0.95.
    At each query 95% of the instances would be sampled with the maximum
    query strategy after which the remaining 5% would be sampled with
    the random query strategy.
    """

    name = "max_random"
    label = "Mixed (95% Maximum and 5% Random)"

    def __init__(self,
                 mix_ratio=0.95,
                 random_state=None,
                 **kwargs):
        """Initialize the Mixed (Maximum and Random) query strategy."""
        super(MaxRandomQuery, self).__init__(
            strategy_1="max",
            strategy_2="random",
            mix_ratio=mix_ratio,
            random_state=random_state,
            **kwargs)


class MaxUncertaintyQuery(MixedQuery):
    """Mixed (95% Maximum and 5% Uncertainty) query strategy.

    A mix of maximum and random query strategies with a mix ratio of 0.95.
    At each query 95% of the instances would be sampled with the maximum
    query strategy after which the remaining 5% would be sampled with
    the uncertainty query strategy.
    """

    name = "max_uncertainty"
    label = "Mixed (95% Maximum and 5% Uncertainty)"

    def __init__(self,
                 mix_ratio=0.95,
                 random_state=None,
                 **kwargs):
        """Initialize the Mixed (Maximum and Uncertainty) query strategy."""
        super(MaxUncertaintyQuery, self).__init__(
            strategy_1="max",
            strategy_2="uncertainty",
            mix_ratio=mix_ratio,
            random_state=random_state,
            **kwargs)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import logging
import sys

import numpy as np
import pandas as pd

from asreview.compat import convert_id_to_idx, convert_idx_to_id
from asreview.review.factory import get_reviewer
from asreview.state.utils import open_state
from asreview.webapp.sqlock import SQLiteLock
from asreview.webapp.utils import get_lock_path
from asreview.webapp.utils import get_state_path
from asreview.webapp.utils.io import read_label_history
from asreview.webapp.utils.io import read_pool
from asreview.webapp.utils.io import write_pool
from asreview.webapp.utils.io import write_proba
from asreview.webapp.utils.paths import get_data_file_path
from asreview.webapp.utils.paths import get_project_path
from asreview.webapp.utils.paths import get_kwargs_path
from asreview.webapp.utils.project import read_data


def _get_diff_history(new_history, old_history):
    for i in range(len(new_history)):
        try:
            if old_history[i] != new_history[i]:
                return new_history[i:]
        except IndexError:
            return new_history[i:]
    return []


def _get_label_train_history(state):
    label_idx = []
    inclusions = []
    for query_i in range(state.n_queries()):
        try:
            new_labels = state.get("label_idx", query_i=query_i)
            new_inclusions = state.get("inclusions", query_i=query_i)
        except KeyError:
            new_labels = None
        if new_labels is not None:
            label_idx.extend(new_labels)
            inclusions.extend(new_inclusions)

    return list(zip(label_idx, inclusions))


def train_model(project_id, label_method=None):
    """Add the new labels to the review and do the modeling.

    It uses a lock to ensure only one model is running at the same time.
    Old results directories are deleted after 4 iterations.

    It has one argument on the CLI, which is the base project directory.
    """

    logging.info(f"Project {project_id} - Train a new model for project")

    # get file locations
    asr_kwargs_file = get_kwargs_path(project_id)
    lock_file = get_lock_path(project_id)

    # Lock so that only one training run is running at the same time.
    # It doesn't lock the flask server/client.
    with SQLiteLock(
            lock_file, blocking=False, lock_name="training",
            project_id=project_id) as lock:

        # If the lock is not acquired, another training instance is running.
        if not lock.locked():
            logging.info("Project {project_id} - "
                         "Cannot acquire lock, other instance running.")
            return

        # Lock the current state. We want to have a consistent active state.
        # This does communicate with the flask backend; it prevents writing and
        # reading to the same files at the same time.
        with SQLiteLock(
                lock_file,
                blocking=True,
                lock_name="active",
                project_id=project_id) as lock:
            # Get the all labels since last run. If no new labels, quit.
            new_label_history = read_label_history(project_id)

        data_fp = str(get_data_file_path(project_id))
        as_data = read_data(project_id)
        state_file = get_state_path(project_id)

        # collect command line arguments and pass them to the reviewer
        with open(asr_kwargs_file, "r") as fp:
            asr_kwargs = json.load(fp)

        try:
            del asr_kwargs["abstract_only"]
        except KeyError:
            pass

        asr_kwargs['state_file'] = str(state_file)
        reviewer = get_reviewer(dataset=data_fp, mode="minimal", **asr_kwargs)

        with open_state(state_file) as state:
            old_label_history = _get_label_train_history(state)

        diff_history = _get_diff_history(new_label_history, old_label_history)

        if len(diff_history) == 0:
            logging.info(
                "Project {project_id} - No new labels since last run.")
            return

        query_record_ids = np.array([x[0] for x in diff_history], dtype=int)
        inclusions = np.array([x[1] for x in diff_history], dtype=int)

        query_idx = convert_id_to_idx(as_data, query_record_ids)

        # Classify the new labels, train and store the results.
        with open_state(state_file) as state:
            reviewer.classify(
                query_idx, inclusions, state, method=label_method)
            reviewer.train()
            reviewer.log_probabilities(state)
            new_query_idx = reviewer.query(reviewer.n_pool()).tolist()
            reviewer.log_current_query(state)

            # write the proba to a pandas dataframe with record_ids as index
            proba = pd.DataFrame(
                {"proba": state.pred_proba.tolist()},
                index=pd.Index(as_data.record_ids, name="record_id")
            )

        # update the pool and output the proba's
        # important: pool is sorted on query
        with SQLiteLock(
                lock_file,
                blocking=True,
                lock_name="active",
                project_id=project_id) as lock:

            # read the pool
            current_pool = read_pool(project_id)

            # diff pool and new_query_ind
            current_pool_idx = convert_id_to_idx(as_data, current_pool)
            current_pool_idx = frozenset(current_pool_idx)
            new_pool_idx = [x for x in new_query_idx if x in current_pool_idx]

            # convert new_pool_idx back to record_ids
            new_pool = convert_idx_to_id(as_data, new_pool_idx)

            # write the pool and proba
            write_pool(project_id, new_pool)
            write_proba(project_id, proba)


def main(argv):

    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("project_id", type=str, help="Project id")
    parser.add_argument(
        "--label_method",
        type=str,
        default=None,
        help="Label method (for example 'prior')")
    args = parser.parse_args(argv)

    try:
        train_model(args.project_id, args.label_method)
    except Exception as err:
        err_type = type(err).__name__
        logging.error(f"Project {args.project_id} - {err_type}: {err}")

        # write error to file is label method is prior (first iteration)
        if args.label_method == "prior":
            message = {"message": f"{err_type}: {err}"}

            fp = get_project_path(args.project_id) / "error.json"
            with open(fp, 'w') as f:
                json.dump(message, f)

        # raise the error for full traceback
        raise err


if __name__ == "__main__":

    main(sys.argv)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from asreview.entry_points.base import BaseEntryPoint
from asreview.entry_points.base import _base_parser
from asreview.webapp.run_model import main as main_run_model


HOST_NAME = "localhost"
PORT_NUMBER = 5000


def _lab_parser(prog="lab"):
    parser = _base_parser(
        prog=prog,
        description="""ASReview LAB - Active learning for Systematic Reviews."""  # noqa
    )

    parser.add_argument(
        "--clean-project",
        dest="clean_project",
        default=None,
        type=str,
        help="Safe cleanup of temporary files in project.")

    parser.add_argument(
        "--clean-all-projects",
        dest="clean_all_projects",
        default=None,
        action='store_true',
        help="Safe cleanup of temporary files in all projects.")

    parser.add_argument(
        "--ip",
        default=HOST_NAME,
        type=str,
        help="The IP address the server will listen on.")

    parser.add_argument(
        "--port",
        default=PORT_NUMBER,
        type=int,
        help="The port the server will listen on.")

    parser.add_argument(
        "--no-browser",
        dest="no_browser",
        action='store_true',
        help="Do not open ASReview LAB in a browser after startup.")

    parser.add_argument(
        "--port-retries",
        dest="port_retries",
        default=50,
        type=int,
        help="The number of additional ports to try if the"
        "specified port is not available.")

    parser.add_argument(
        "--certfile",
        default="",
        type=str,
        help="The full path to an SSL/TLS certificate file.")

    parser.add_argument(
        "--keyfile",
        default="",
        type=str,
        help="The full path to a private key file for usage with SSL/TLS.")

    return parser


class LABEntryPoint(BaseEntryPoint):
    description = "Graphical user interface for ASReview."

    def execute(self, argv):

        from asreview.webapp.start_flask import main

        main(argv)


# deprecated oracle class
class OracleEntryPoint(LABEntryPoint):
    description = "Graphical user interface for ASReview. (Deprecated)"

    def execute(self, argv):

        logging.warning("Warning: subcommmand 'oracle' is replaced by 'lab'.")

        super(OracleEntryPoint, self).execute(argv)


class WebRunModelEntryPoint(BaseEntryPoint):
    description = "Internal use only."

    def execute(self, argv):
        main_run_model(argv)
